/* 28-30 January 2012, copyright R.Harmsen:

   Update the info about sizes, weight and price of gold at
   http://rudhar.com/gold automatically, based on a date and
   price per ounce in dollars and euros.

   8 June 2015, added support for Interlingua.

   10 January 2022, first check if an update is necessary.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#ifndef TestThousandsSpaces

#if defined __MSDOS__ || defined _WIN32
   #define PathPrefix "d:/rudhar.com"
#else
// #define PathPrefix "/usr/local/www/apache2/htdocs"
// #define PathPrefix "/usr/local/www/apache22/htdocs"
// #define PathPrefix "/usr/local/www/apache24/htdocs"
   #define PathPrefix "/var/www/html"
#endif

long DateOfPriceYYYYMMDD;
double DollarPerOunce, EuroPerOunce;

static char *NameDataFile = "./updglddt";
static struct FilesToHandle_s
{
   char PathPostfix[31+1];
   char Language[2+1];
   char SearchString[31+1];
} FileToHandle[] =
{
   { "gold/ar.stm", "ar", ">\xD8\xA3\xD8\xB3\xD8\xB9\xD8\xA7\xD8\xB1\x20\xD8\xA7\xD9\x84\xD8\xB0\xD9\x87\xD8\xA8</a>"},
   { "gold/de.stm", "de", ">Goldpreisen</a>"},
   { "gold/en.stm", "en", ">gold prices</a>"},
   { "gold/es.stm", "es", ">precios de oro</a>"},
   { "gold/fr.stm", "fr", ">cours de l’or</a>"},
   { "gold/ia.stm", "ia", ">precios del auro</a>"},
   { "gold/nl.stm", "nl", ">goudprijzen</a>"},
   { "gold/pt.stm", "pt", ">preços de ouro</a>"},
   { "gold/tablbody.inc"   , ""  , ""},
};

static int NecessaryAtAll (void);
static int ReadInputData (long *p_DateOfPriceYYYYMMDD,
   double *p_DollarPerOunce, double *p_EuroPerOunce);
static int HandleThatFile (int elt);
static int DoLanguageFile (int elt, char *PathBuf, char *FopenMode);
static int DoUpdate (FILE *fp, int elt);
static long FindNextDigit (FILE *fp, int reversed);
static int DoTablbody (char *PathBuf, char *FopenMode);
static char *ThousandsSpaces (double number, int numdecimals);
static char *ProgName;

int main (int argc, char **argv)
{
   int elt;

   ProgName = argv[0];

   if (!NecessaryAtAll())
      return 0;

   /* First get the three data items from an input file.
      Example of its contents:
      20111114 1782.18 1304.50 */
   if (ReadInputData(&DateOfPriceYYYYMMDD, &DollarPerOunce, &EuroPerOunce))
   {
       fprintf(stderr, "%s: No input data, program does nothing\n", argv[0]);
       return 1;
   }

   for (elt = 0; elt < (sizeof FileToHandle) / (sizeof FileToHandle[0]); elt++)
   {
      HandleThatFile(elt);
   }

   return 0;
}

static int NecessaryAtAll (void)
{
   struct stat dataf, htmlf;
   char PathBuf[133];

   memset(&dataf, '\0', sizeof dataf);
   memset(&htmlf, '\0', sizeof htmlf);
   sprintf(PathBuf, "%s/%s", PathPrefix, FileToHandle[0].PathPostfix);
   stat(NameDataFile, &dataf);
   stat(PathBuf     , &htmlf);

   if (dataf.st_mtime > htmlf.st_mtime)
      return 1;
   return 0;
}

static int ReadInputData (long *p_DateOfPriceYYYYMMDD,
   double *p_DollarPerOunce, double *p_EuroPerOunce)
{
   FILE *fp;
   char buf[81];

   fp = fopen(NameDataFile, "r");
   if (!fp)
   {
      fprintf(stderr, "Can't open datafile %s for reading\n", NameDataFile);
      return 1;
   }

   fread(buf, (sizeof buf) - 1, 1, fp);
   sscanf(buf, "%ld %lf %lf",
      p_DateOfPriceYYYYMMDD, p_DollarPerOunce, p_EuroPerOunce);

   fclose(fp);

   return 0;
}

static int HandleThatFile (int elt)
{
   char PathBuf[133];

   sprintf(PathBuf, "%s/%s", PathPrefix, FileToHandle[elt].PathPostfix);

   if (FileToHandle[elt].Language[0] == ' ' ||
       FileToHandle[elt].Language[0] == '\0')
   {
      DoTablbody(PathBuf, "w+b");
   }
   else
   {
      DoLanguageFile(elt, PathBuf, "r+b");
   }

   return 0;
}

static int DoLanguageFile (int elt, char *PathBuf, char *FopenMode)
{
   FILE *fp;
   char Buf[133];

   fp = fopen(PathBuf, FopenMode);
   if (!fp)
   {
      fprintf(stderr, "%s: Cannot open %s, mode %s\n", ProgName, PathBuf, FopenMode);
      perror(NULL);
      return 1;
   }

   while (fgets(Buf, sizeof Buf, fp) != NULL)
   {
      if (strstr(Buf, FileToHandle[elt].SearchString) != NULL)
      {
         DoUpdate(fp, elt);
         break;
      }
   }
   fclose(fp);

   return 0;
}

static int DoUpdate (FILE *fp, int elt)
{
   char DataBuf[2+1 + 2+1 + 4+1];
   int DateSepChar = '-';
   int Day, Month, Year;
   char *p;

   if (FindNextDigit(fp, 0) == EOF)
      return 1;

   /* Now in the right position to update the date in the file
      (update in place) */
   if (strcmp(FileToHandle[elt].Language, "de") == 0)
      DateSepChar = '.';
   else if (strcmp(FileToHandle[elt].Language, "en") == 0)
      DateSepChar = '/';

   Day   =  DateOfPriceYYYYMMDD % 100;
   Month = (DateOfPriceYYYYMMDD / 100) % 100;
   Year  = (DateOfPriceYYYYMMDD / 10000) % 10000;

   if (strcmp(FileToHandle[elt].Language, "en") == 0)
      sprintf(DataBuf, "%02d%c%02d%c%04d", Month, DateSepChar, Day, DateSepChar, Year);
   else if (strcmp(FileToHandle[elt].Language, "ar") == 0)
      sprintf(DataBuf, "%04d%c%02d%c%02d", Year, DateSepChar, Month, DateSepChar, Day);
   else
      sprintf(DataBuf, "%02d%c%02d%c%04d", Day, DateSepChar, Month, DateSepChar, Year);

   fwrite(DataBuf, strlen(DataBuf), 1, fp);
   /* Intermediate fseek between write and next read, some
      stdio libraries need this? */
   fseek(fp, 0, SEEK_CUR);

   if (FindNextDigit(fp, 0) == EOF)
      return 1;

   /* And update the gold prices */
   sprintf(DataBuf, "%-7.2f", DollarPerOunce);

   /* Decimal point is comma for languages other than English */
   if (strcmp(FileToHandle[elt].Language, "en") != 0)
   {
      if ((p = strchr(DataBuf, '.')) != NULL)
         *p =',';
   }

   fwrite(DataBuf, strlen(DataBuf), 1, fp);
   fseek(fp, 0, SEEK_CUR);

   /* Just for safety, skip to the next non-digit character,
      in case there was an extra decimal in the original file,
      like there once used to be in the arabic file. */
   if (FindNextDigit(fp, 1) == EOF)
      return 1;

   if (FindNextDigit(fp, 0) == EOF)
      return 1;

   sprintf(DataBuf, "%-7.2f", EuroPerOunce);

   /* Decimal point is comma for languages other than English */
   if (strcmp(FileToHandle[elt].Language, "en") != 0)
   {
      if ((p = strchr(DataBuf, '.')) != NULL)
         *p =',';
   }

   fwrite(DataBuf, strlen(DataBuf), 1, fp);
   fseek(fp, 0, SEEK_CUR);

   return 0;
}

/* Returns file postition of next digit (file pointer
   also set to that point), or EOF for end-of-file = not-found.
*/
static long FindNextDigit (FILE *fp, int reversed)
{
   int c;

   while (c = getc(fp), 1)
   {
      if (c == EOF)
      {
         return EOF;
      }
      else if ((!reversed && isdigit(c)) || (reversed && !isdigit(c)))
      {
         fseek(fp, -1, SEEK_CUR);
         return ftell(fp);
      }
   }
   return EOF;
}

static int DoTablbody (char *PathBuf, char *FopenMode)
{
   FILE *fp;
   const double KilogramsInOneGrain = 0.45359237; /* B27 in the spreadsheet */
   const double PoundsInOneGrain = 480.0/7000.0;  /* A12 in the spreadsheet */
   const float  KilogramsIn1dm3OfPureGold = 19.2; /* B28 in the spreadsheet */
   const float  MillimetersInOneInch      = 25.4; /* B26 in the spreadsheet */

   fp = fopen(PathBuf, FopenMode);
   if (!fp)
   {
      fprintf(stderr, "%s: Cannot open %s, mode %s\n", ProgName, PathBuf, FopenMode);
      return 1;
   }

   fprintf(fp, "<tbody>\n");

   {
      double H7, I7;

      /* Exactly one grain of gold */
      fprintf(fp, "<tr>\n");
      fprintf(fp, "  <td class=rite>1/7000</td> <td class=rite>1/5760</td> <td class=rite>1/480</td> <td class=rite>1.000</td>");
      H7 = (1000.0 / KilogramsIn1dm3OfPureGold / 1000.0) * (KilogramsInOneGrain / 7000.0 * 1000.0);
      I7 = pow(H7 * 1000, 1.0/(double)3);
      fprintf(fp, " <td class=rite>%.3f</td> <td class=rite>%.2f</td> <td class=rite>%.2f</td> <td class=rite>%.4f</td> <td class=rite>%.3f</td> <td class=rite>%.3f</td>",
         KilogramsInOneGrain / 7000.0 * 1000.0,
         DollarPerOunce / 480.0,
         EuroPerOunce / 480.0,
         H7, I7, I7 / MillimetersInOneInch);
      fprintf(fp, "\n</tr>\n");
   }

   {
      double D8, H8, I8;

      /* Exactly one metric gram of gold */
      fprintf(fp, "<tr>\n");
      D8 = 7000.0 / KilogramsInOneGrain / 1000.0;
      fprintf(fp, "  <td class=rite>%.5f</td> <td class=rite>%.5f</td> <td class=rite>%.3f</td> <td class=rite>%.3f</td> <td class=rite>1.000</td>",
         D8 / 7000.0, D8 / 5760.0, D8 / 480.0, D8);
      H8 = (1000.0 / KilogramsIn1dm3OfPureGold / 1000.0);
      I8 = pow(H8 * 1000, 1.0/(double)3);
      fprintf(fp, " <td class=rite>%.2f</td> <td class=rite>%.2f</td> <td class=rite>%.4f</td> <td class=rite>%.3f</td> <td class=rite>%.3f</td>",
         DollarPerOunce * 7000.0 / KilogramsInOneGrain / 480.0 / 1000.0,
         EuroPerOunce   * 7000.0 / KilogramsInOneGrain / 480.0 / 1000.0,
         H8, I8, I8 / MillimetersInOneInch);
      fprintf(fp, "\n</tr>\n");
   }

   {
      double F9, C9, D9, E9, H9, I9;

      /* Exactly one hundred dollars */
      fprintf(fp, "<tr>\n");
      F9 = 100.00;
      C9 = F9 / DollarPerOunce;
      D9 = C9 * 480.0;
      E9 = F9 * KilogramsInOneGrain * 100.0 * 4800.0 / DollarPerOunce / 7000.0;
      H9 = (F9 * KilogramsInOneGrain * 1000.0 * 480.0) / (DollarPerOunce * KilogramsIn1dm3OfPureGold * 7000.0);
      I9 = pow(H9 * 1000, 1.0/(double)3);
      fprintf(fp, "  <td class=rite>%.5f</td> <td class=rite>%.5f</td> <td class=rite>%.3f</td> <td class=rite>%.3f</td> <td class=rite>%.3f</td>",
         D9 / 7000.0, D9 / 5760.0, C9, D9, E9);
      fprintf(fp, " <td class=rite>%.2f</td> <td class=rite>%.2f</td> <td class=rite>%.3f</td> <td class=rite>%.3f</td> <td class=rite>%.3f</td>",
         F9, F9 * EuroPerOunce / DollarPerOunce,
         H9, I9, I9 / MillimetersInOneInch);
      fprintf(fp, "\n</tr>\n");
   }

   {
      double G10, C10, D10, E10, H10, I10;

      /* Exactly one hundred euros */
      fprintf(fp, "<tr>\n");
      G10 = 100.00;
      C10 = G10 / EuroPerOunce;
      D10 = C10 * 480.0;
      E10 = G10 * KilogramsInOneGrain * 100.0 * 4800.0 / EuroPerOunce / 7000.0;
      H10 = (G10 * KilogramsInOneGrain * 1000.0 * 480.0) / (EuroPerOunce * KilogramsIn1dm3OfPureGold * 7000.0);
      I10 = pow(H10 * 1000, 1.0/(double)3);
      fprintf(fp, "  <td class=rite>%.5f</td> <td class=rite>%.5f</td> <td class=rite>%.3f</td> <td class=rite>%.3f</td> <td class=rite>%.3f</td>",
         D10 / 7000.0, D10 / 5760.0, C10, D10, E10);
      fprintf(fp, " <td class=rite>%.2f</td> <td class=rite>%.2f</td> <td class=rite>%.3f</td> <td class=rite>%.3f</td> <td class=rite>%.3f</td>",
         G10 * DollarPerOunce / EuroPerOunce, G10,
         H10, I10, I10 / MillimetersInOneInch);
      fprintf(fp, "\n</tr>\n");
   }

   {
      /* Exactly one cubic centimeter */
      double D11, H11, I11;

      fprintf(fp, "<tr>\n");
      D11 = KilogramsIn1dm3OfPureGold / KilogramsInOneGrain * (7000.0 / 1000.0);
      fprintf(fp, "  <td class=rite>%.5f</td> <td class=rite>%.5f</td> <td class=rite>%.3f</td> <td class=rite>%.3f</td> <td class=rite>%.3f</td>",
         D11 / 7000.0, D11 / 5760.0, D11 / 480.0, D11, KilogramsIn1dm3OfPureGold);
      H11 = 1.0;
      I11 = pow(H11 * 1000, 1.0/(double)3);
      fprintf(fp, " <td class=rite>%s</td>",
         ThousandsSpaces((KilogramsIn1dm3OfPureGold * DollarPerOunce * 7000.0) / (KilogramsInOneGrain * 480.0 * 1000.0), 2));
      fprintf(fp, " <td class=rite>%s</td>",
         ThousandsSpaces((KilogramsIn1dm3OfPureGold *   EuroPerOunce * 7000.0) / (KilogramsInOneGrain * 480.0 * 1000.0), 2));
      fprintf(fp, " <td class=rite>%.3f</td> <td class=rite>%.1f</td> <td class=rite>%.3f</td>",
         H11, I11, I11 / MillimetersInOneInch);
      fprintf(fp, "\n</tr>\n");
   }

   {
      /* Exactly one Troy ounce */
      double H12, I12;

      fprintf(fp, "<tr>\n");
      fprintf(fp, "  <td class=rite>%.5f</td> <td class=rite>1/12</td> <td class=rite>1.000</td> <td class=rite>480</td> <td class=rite>%.3f</td>",
         PoundsInOneGrain, KilogramsInOneGrain * 480.0 * 1000.0 / 7000.00);

      fprintf(fp, " <td class=rite>%s</td>", ThousandsSpaces(DollarPerOunce, 2));
      fprintf(fp, " <td class=rite>%s</td>", ThousandsSpaces(EuroPerOunce  , 2));

      H12 = KilogramsInOneGrain * 1000.0 * 480.0 / (KilogramsIn1dm3OfPureGold * 7000.0);
      I12 = pow(H12 * 1000, 1.0/(double)3);
      fprintf(fp, " <td class=rite>%.3f</td> <td class=rite>%.1f</td> <td class=rite>%.3f</td>",
         H12, I12, I12 / MillimetersInOneInch);
      fprintf(fp, "\n</tr>\n");
   }

   {
      /* Exactly one cubic inch */
      double D13, E13, powpow, factor, F13, G13, H13, I13, J13;

      fprintf(fp, "<tr>\n");
      powpow = pow(MillimetersInOneInch, 3.0);
      D13 = KilogramsIn1dm3OfPureGold * 7000.0 * powpow / KilogramsInOneGrain / 1000.0 / 1000.0;
      E13 = KilogramsIn1dm3OfPureGold * powpow / 1000.0;
      fprintf(fp, "  <td class=rite>%.3f</td> <td class=rite>%.3f</td> <td class=rite>%.3f</td> <td class=rite>%s</td> <td class=rite>%.0f</td>",
         D13 / 7000.0, D13 / 5760.0, D13 / 480.0,
         ThousandsSpaces(D13, 0),
         E13);
      factor = powpow * KilogramsIn1dm3OfPureGold * 7000 / (KilogramsInOneGrain * 480 * 1000 * 1000);
      F13 = DollarPerOunce * factor;
      G13 =   EuroPerOunce * factor;
      J13 = 1.0;
      I13 = MillimetersInOneInch * J13;
      H13 = pow(I13 / 10.0, 3.0);
      fprintf(fp, " <td class=rite>%s</td>", ThousandsSpaces(F13, 0));
      fprintf(fp, " <td class=rite>%s</td>", ThousandsSpaces(G13, 0));
      fprintf(fp, " <td class=rite>%.3f</td> <td class=rite>%.1f</td> <td class=rite>%.3f</td>", H13, I13, J13);
      fprintf(fp, "\n</tr>\n");
   }

   {
      /* Exactly one Troy pound */
      double H14, I14;

      fprintf(fp, "<tr>\n");
      fprintf(fp, "  <td class=rite>%.3f</td> <td class=rite>1.000</td> <td class=rite>12.000</td> <td class=rite>%s</td> <td class=rite>%.0f</td>",
         12 * PoundsInOneGrain,
         ThousandsSpaces(5760.0, 0),
         12 * KilogramsInOneGrain * 480.0 * 1000.0 / 7000.00);

      fprintf(fp, " <td class=rite>%s</td>", ThousandsSpaces(DollarPerOunce * 12.0, 0));
      fprintf(fp, " <td class=rite>%s</td>", ThousandsSpaces(EuroPerOunce   * 12.0, 0));

      H14 = KilogramsInOneGrain * 1000.0 * 5760.0 / (KilogramsIn1dm3OfPureGold * 7000.0);
      I14 = pow(H14 * 1000, 1.0/(double)3);
      fprintf(fp, " <td class=rite>%.3f</td> <td class=rite>%.1f</td> <td class=rite>%.3f</td>",
         H14, I14, I14 / MillimetersInOneInch);
      fprintf(fp, "\n</tr>\n");
   }

   {
      /* Exactly one metric kilogram of gold */
      double D15, H15, I15;

      fprintf(fp, "<tr>\n");
      D15 = 7000.0 / KilogramsInOneGrain;

      fprintf(fp, "  <td class=rite>%.3f</td> <td class=rite>%.3f</td> <td class=rite>%.3f</td> <td class=rite>%s</td>",
         D15 / 7000.0, D15 / 5760.0, D15 / 480.0,
         ThousandsSpaces(D15, 0));
      fprintf(fp, " <td class=rite>%s</td>",
         ThousandsSpaces(1000.0, 0));
      fprintf(fp, " <td class=rite>%s</td>",
         ThousandsSpaces(DollarPerOunce * 7000.0 / KilogramsInOneGrain / 480.0, 0));
      fprintf(fp, " <td class=rite>%s</td>",
         ThousandsSpaces(EuroPerOunce   * 7000.0 / KilogramsInOneGrain / 480.0, 0));
      H15 = 1000.0 / KilogramsIn1dm3OfPureGold;
      I15 = pow(H15 * 1000, 1.0/(double)3);
      fprintf(fp, " <td class=rite>%.3f</td> <td class=rite>%.1f</td> <td class=rite>%.3f</td>",
         H15, I15, I15 / MillimetersInOneInch);
      fprintf(fp, "\n</tr>\n");
   }

   {
      /* Exactly 400 Troy ounce, the so-called bullion (Dutch: broodje) */
      double H16, I16;

      fprintf(fp, "<tr>\n");
      fprintf(fp, "  <td class=rite>%.3f</td> <td class=rite>%.3f</td> <td class=rite>%.3f</td> <td class=rite>%s</td>",
         400 * PoundsInOneGrain,
         400 / 12.0,
         400.0,
         ThousandsSpaces(400.0 * 480, 0));

      fprintf(fp, "  <td class=rite>%s</td>",
         ThousandsSpaces(400 * KilogramsInOneGrain * 480.0 * 1000.0 / 7000.00, 0));

      fprintf(fp, " <td class=rite>%s</td>", ThousandsSpaces(DollarPerOunce * 400.0, 0));
      fprintf(fp, " <td class=rite>%s</td>", ThousandsSpaces(EuroPerOunce   * 400.0, 0));

      H16 = KilogramsInOneGrain * 1000.0 * 480.0 / (KilogramsIn1dm3OfPureGold * 7000.0) * 400.0;
      I16 = pow(H16 * 1000, 1.0/(double)3);
      fprintf(fp, " <td class=rite>%.0f</td> <td class=rite>%.1f</td> <td class=rite>%.3f</td>",
         H16, I16, I16 / MillimetersInOneInch);
      fprintf(fp, "\n</tr>\n");
   }

   {
      /* Exactly one cubic decimeter = one liter */
      double D17, H17, I17;

      fprintf(fp, "<tr>\n");
      D17 = KilogramsIn1dm3OfPureGold / KilogramsInOneGrain * 7000.0;
      fprintf(fp, "  <td class=rite>%.3f</td> <td class=rite>%.3f</td> <td class=rite>%.3f</td> <td class=rite>%s</td>",
         D17 / 7000.0, D17 / 5760.0, D17 / 480.0,
         ThousandsSpaces(D17, 0));
      fprintf(fp, " <td class=rite>%s</td>", ThousandsSpaces(KilogramsIn1dm3OfPureGold * 1000, 0));
      H17 = 1000.0;
      I17 = pow(H17 * 1000, 1.0/(double)3);
      fprintf(fp, " <td class=rite>%s</td>",
         ThousandsSpaces((KilogramsIn1dm3OfPureGold * DollarPerOunce * 7000.0) / (KilogramsInOneGrain * 480.0), 0));
      fprintf(fp, " <td class=rite>%s</td>",
         ThousandsSpaces((KilogramsIn1dm3OfPureGold *   EuroPerOunce * 7000.0) / (KilogramsInOneGrain * 480.0), 0));
      fprintf(fp, " <td class=rite>%.0f</td> <td class=rite>%.1f</td> <td class=rite>%.3f</td>",
         H17, I17, I17 / MillimetersInOneInch);
      fprintf(fp, "\n</tr>\n");
   }

   fprintf(fp, "</tbody>\n");

   fclose(fp);

   return 0;
}

/* This endif refer to an #ifndef TestThousandsSpaces at the beginning of
   this source file */
#endif


/* Groups digits by three, with a (thin) space in between.
   Returns a pointer to a static buffer. */

#define ThousandsSpacesRequiredSpace (5*3 + 5*8 + 3)
static char *ThousandsSpaces (double number, int numdecimals)
{
   // Thin space disabled, because they cannot be defined as non-breaking AFAIK
   // const char *SpaceCode = "&#x200A;";
   // const char *SpaceCode = "&nbsp;";
   // But this works well: U+202F Narrow No-Break Space
   const char *SpaceCode = "&#x202F;";

   static char retbuf[ThousandsSpacesRequiredSpace];
   char wrkbuf[ThousandsSpacesRequiredSpace];
   int DigitsBeforeDecimalPoint;
   int DecimalSeparators;
   int NeededSpace;
   int i, pos_s = 0, pos_t = 0;

   sprintf(wrkbuf, "%.*f", numdecimals, number);
   DigitsBeforeDecimalPoint = strlen(wrkbuf);
   if (numdecimals > 0)
   {
      DigitsBeforeDecimalPoint -= (numdecimals + 1);
   }
   if (number < 0.0)
   {
      /* Account for the minus sign */
      DigitsBeforeDecimalPoint--;
   }

   DecimalSeparators = (DigitsBeforeDecimalPoint - 1) / 3;

   /* Safety fist, avoid buffer overflow */
   NeededSpace = strlen(wrkbuf) + DecimalSeparators * strlen(SpaceCode) + 1;
   if (NeededSpace > sizeof retbuf)
   {
      sprintf(retbuf, "Oflow");
      return retbuf;
   }

   /* Safety first */
   //memset(retbuf, '\0', sizeof retbuf);
   memset(retbuf, ' ', sizeof retbuf);
   retbuf[(sizeof retbuf) - 1] = '\0';

   if (DigitsBeforeDecimalPoint > DecimalSeparators * 3)
   {
      int adv;

      adv = ((number < 0.0) ? 1 : 0) + DigitsBeforeDecimalPoint - DecimalSeparators * 3;
      sprintf(retbuf, "%.*s", adv, wrkbuf);

      pos_s += adv;
      pos_t += adv;
   }
   for (i = 0; i < DecimalSeparators; i++)
   {
      sprintf(retbuf + pos_t, "%s%.3s", SpaceCode, wrkbuf + pos_s);
      pos_s += 3;
      pos_t += strlen(SpaceCode) + 3;
   }

   if (numdecimals > 0)
   {
      sprintf(retbuf + DigitsBeforeDecimalPoint + ((number < 0.0) ? 1 : 0) + DecimalSeparators * strlen(SpaceCode),
         "%.*s", 1 + numdecimals,
         wrkbuf + DigitsBeforeDecimalPoint + ((number < 0.0) ? 1 : 0));

      retbuf[DigitsBeforeDecimalPoint + ((number < 0.0) ? 1 : 0) + DecimalSeparators * strlen(SpaceCode)] = '.';
   }

   return retbuf;
}

#ifdef TestThousandsSpaces

static double TestTab[] =
{
           -1234567890.123,
                -12234.12,
                   -12.0909,
                    -1.1234567,
                     1.1234567,
                    12.0909,
                   123,
                  1000,
                  9999.12,
                 70090.98,
                123456.77,
               1234567.92828,
          123456789012.12,
   1234567890123456789.1234,
1090909090909009090909.1245,
};

int main (int argc, char **argv)
{
   int i;

   for (i = 0; i < sizeof TestTab / sizeof TestTab[0]; i++)
   {
      printf("\t(%2d): %28f\t:%s:\t", i, TestTab[i],
         ThousandsSpaces(TestTab[i], 0));
      printf(":%s:\n\n",
         ThousandsSpaces(TestTab[i], 2));
   }
   return 0;
}

#endif
